/*
Copyright (C) 2002 Observatoire de Paris-Meudon

Ce programme est un logiciel libre ; vous pouvez le redistribuer et/ou le modifier conformment aux dispositions de la Licence Publique Gnrale GNU, telle que publie par la Free Software Foundation ; version 2 de la licence, ou encore ( votre choix) toute version ultrieure.

Ce programme est distribu dans l'espoir qu'il sera utile, mais SANS AUCUNE GARANTIE ; sans mme la garantie implicite de COMMERCIALISATION ou D'ADAPTATION A UN OBJET PARTICULIER. Pour plus de dtail, voir la Licence Publique Gnrale GNU .

Vous devez avoir reu un exemplaire de la Licence Publique Gnrale GNU en mme temps que ce programme ; si ce n'est pas le cas, crivez  la Free Software Foundation Inc., 675 Mass Ave, Cambridge, MA 02139, Etats-Unis.
*/

import java.awt.*;
import java.awt.event.*;
import java.applet.*;
import java.awt.image.*;
import java.util.ResourceBundle;

/**
 * Affichage d'une image -> Coupe -> affichage du graphe
 */
public class Coupes extends Applet implements ActionListener {

    static ResourceBundle rb  = ResourceBundle.getBundle("LocalStrings");
    ImageCanvas ic;
    Label lab;
    Image im;
    String labelx, labely;
    double cx1, cy1;
    
    public void init() {
        String nomImage = getParameter("image");
        String sx1 = getParameter("x1");
        String sx2 = getParameter("x2");
        String sy1 = getParameter("y1");
        String sy2 = getParameter("y2");
        labelx = getParameter("labelx");
        if (sx1 == null || sx2 == null)
            labelx = null;
        else if (labelx == null)
            labelx = "x";
        labely = getParameter("labely");
        if (sy1 == null || sy2 == null)
            labely = null;
        else if (labely == null)
            labely = "y";
        if  (nomImage == null || ((sx1 == null || sx2 == null) && (sy1 == null || sy2 == null))) {
            System.err.println("Il manque un des paramtres de l'applet");
            return;
        }
        double x1,x2,y1,y2;
        try {
            if (labelx == null)
                x1 = x2 = 0;
            else {
                x1 = (new Double(sx1)).doubleValue();
                x2 = (new Double(sx2)).doubleValue();
            }
            if (labely == null)
                y1 = y2 = 0;
            else {
                y1 = (new Double(sy1)).doubleValue();
                y2 = (new Double(sy2)).doubleValue();
            }
        } catch (NumberFormatException ex) {
            System.err.println("Coordonnes mal formattes.");
            return;
        }
        cx1 = x1;
        cy1 = y1;
        
        setLayout(new BorderLayout());
        Panel impane = new Panel(new BorderLayout());
        im = getImage(getCodeBase(), nomImage);
        ic = new ImageCanvas(im, this, 200, 200, x1, x2, y1, y2, labelx, labely);
        impane.add(ic, BorderLayout.CENTER);
        lab = new Label(" ");
        impane.add(lab, BorderLayout.SOUTH);
        add(impane, BorderLayout.CENTER);
        Panel bpane = new Panel(new FlowLayout());
        Button bgraphe = new Button(rb.getString("Graphe"));
        bgraphe.addActionListener(this);
        bgraphe.setActionCommand("graphe");
        bpane.add(bgraphe);
        add(bpane, BorderLayout.SOUTH);
        validate();
        ic.requestFocus();
    }
    
    public void actionPerformed(ActionEvent ae) {
        String cmd = ae.getActionCommand();
        if ("graphe".equals(cmd))
            graphe();
    }
    
    public void graphe() {
        Point p1 = ic.getP1();
        Point p2 = ic.getP2();
        double ff;
        if (labely == null)
            ff = ic.getFacteurX();
        else if (labelx == null)
            ff = ic.getFacteurY();
        else
            ff = (ic.getFacteurX() + ic.getFacteurY()) / 2;
        int w = im.getWidth(null);
        int h = im.getHeight(null);
        int[] pixels = new int[w * h];
        PixelGrabber pg = new PixelGrabber(im, 0, 0, w, h, pixels, 0, w);
        try {
            pg.grabPixels();
        } catch (InterruptedException e) {
            System.err.println("interrupted waiting for pixels!");
            return;
        }
        if ((pg.getStatus() & ImageObserver.ABORT) != 0) {
            System.err.println("image fetch aborted or errored");
            return;
        }
        int x1 = p1.x;
        int x2 = p2.x;
        int y1 = p1.y;
        int y2 = p2.y;
        int dx = Math.abs(x2 - x1);
        int dy = Math.abs(y2 - y1);
        double[] vx;
        double[] vy;
        int x, y;
        int rgb;
        if (dx > dy) {
            vx = new double[dx];
            vy = new double[dx];
            int sgn;
            if (x2 > x1)
                sgn = 1;
            else
                sgn = -1;
            for (int i=0; i<dx; i++) {
                x = x1 + sgn * i;
                y = y1 + (int)Math.round((y2 - y1) * (((double)i) / dx));
                if (labely == null)
                    vx[i] = cx1 + x * ff;
                else if (labelx == null)
                    vx[i] = cy1 + y * ff;
                else
                    vx[i] = Math.sqrt((x - x1)*(x - x1) + (y - y1)*(y - y1)) * ff;
                rgb = pixels[y * w + x];
                int red   = (rgb >> 16) & 0xff;
                int green = (rgb >>  8) & 0xff;
                int blue  = (rgb      ) & 0xff;
                vy[i] = (red + green + blue) / 3;
            }
        } else {
            vx = new double[dy];
            vy = new double[dy];
            int sgn;
            if (y2 > y1)
                sgn = 1;
            else
                sgn = -1;
            for (int i=0; i<dy; i++) {
                x = x1 + (int)Math.round((x2 - x1) * (((double)i) / dy));
                y = y1 + sgn * i;
                if (labely == null)
                    vx[i] = cx1 + x * ff;
                else if (labelx == null)
                    vx[i] = cy1 + y * ff;
                else
                    vx[i] = Math.sqrt((x - x1)*(x - x1) + (y - y1)*(y - y1)) * ff;
                rgb = pixels[y * w + x];
                int red   = (rgb >> 16) & 0xff;
                int green = (rgb >>  8) & 0xff;
                int blue  = (rgb      ) & 0xff;
                vy[i] = (red + green + blue) / 3.0;
            }
        }
        affGraphe(vx, vy);
    }
    
    class ImageCanvas extends Canvas implements MouseMotionListener, MouseListener, KeyListener {
        Container pappy;
        Image image;
        Dimension size;
        int w, h;
        boolean trueSizeKnown=false;
        MediaTracker tracker;
        Point p1 = null, p2 = null;
        int wleft = 60;
        int hbottom = 50;
        double x1,x2,y1,y2;
        String labelx, labely;
    
        public ImageCanvas(Image image, Container highestContainer, 
                        int initialWidth, int initialHeight,
                        double x1, double x2, double y1, double y2,
                        String labelx, String labely) {
            
            if (image == null) {
                System.err.println("Canvas got invalid image object!");
                return;
            }
    
            this.image = image;
            this.pappy = highestContainer;
            
            w = initialWidth;
            h = initialHeight;
    
            this.x1 = x1;
            this.x2 = x2;
            this.y1 = y1;
            this.y2 = y2;
    
            this.labelx = labelx;
            this.labely = labely;
            
            tracker = new MediaTracker(this);
            tracker.addImage(image, 0);
            try {
                tracker.waitForAll();
            } catch (InterruptedException ex) {
                System.err.println(ex.getClass().getName() + ": " + ex.getMessage());
            }
    
            size = new Dimension(w,h);
            
            setCursor(new Cursor(Cursor.CROSSHAIR_CURSOR));
            addMouseMotionListener(this);
            addMouseListener(this);
            addKeyListener(this);
        }
    
        public Dimension getPreferredSize() {
            return getMinimumSize();
        }
    
        public Dimension getMinimumSize() {
            return size;
        }
    
        public void paint (Graphics g) {
            if (image != null) {
                if (!trueSizeKnown) {
                    int imageWidth = image.getWidth(this);
                    int imageHeight = image.getHeight(this);
    
                    if (tracker.checkAll(true)) {
                        trueSizeKnown = true;
                        if (tracker.isErrorAny()) {
                            System.err.println("Erreur au chargement de l'image");
                        }
                    }
    
                    //Component-initiated resizing.
                    if (((imageWidth > 0) && (w != imageWidth + wleft)) ||
                        ((imageHeight > 0) && (h != imageHeight + hbottom))) {
                        w = imageWidth + wleft;
                        h = imageHeight + hbottom;
                        size = new Dimension(w,h);
                        setSize(w, h);
                        pappy.validate();
                    }
                }
            }
    
            g.drawImage(image, wleft, 0, this);
            g.drawRect(wleft, 0, w - wleft - 1, h - hbottom - 1);
            
            axes(g);
            
            dessinerObjet();
        }
        
        public void axes(Graphics g) {
            Font defonce = g.getFont();
            Font plafond = new Font(defonce.getName(), Font.PLAIN, 10);
            g.setFont(plafond);
            Font boldfont = new Font(defonce.getName(), Font.BOLD, 10);
            
            // Axe x
            
            if (labelx != null) {
                int ndiv = (w - wleft)/30;
                double fx = ndiv/Math.abs(x2 - x1);
                int exposant_fx;
                if (fx != 0)
                    exposant_fx = (int)Math.floor(Math.log(Math.abs(fx))/Math.log(10));
                else
                    exposant_fx = 0;
                double mantisse_fx = fx / (Math.pow(10,exposant_fx));
                mantisse_fx = Math.round(mantisse_fx);
                // on guarde les chiffres dont l'inverse n'a pas trop de dcimales
                if (mantisse_fx == 3)
                    mantisse_fx = 2;
                else if (mantisse_fx == 6)
                    mantisse_fx = 5;
                else if (mantisse_fx > 6)
                    mantisse_fx = 10;
                fx = mantisse_fx * Math.pow(10, exposant_fx);
                grilleX(g, fx);
                /*
                int zoomx = 1;
                if (Math.abs(x2 - x1) < 1.5)
                    zoomx = 2; // affiche les diximes
                if (Math.abs(x2 - x1) < 1.5/10)
                    zoomx = 3; // affiche les centimes
                
                if (zoomx == 1)
                    grilleX(g, 1);
                
                if (zoomx == 2)
                    grilleX(g, 10);
                
                if (zoomx > 2)
                    grilleX(g, 100);
                */
                g.setFont(boldfont);
                String sx = labelx;
                g.drawString(sx, wleft + (w - wleft)/2 - (g.getFontMetrics().stringWidth(sx))/2, h - hbottom/4);
                g.setFont(plafond);
            }
            
            // Axe y
            
            if (labely != null) {
                int ndiv = (h - hbottom)/15;
                double fy = ndiv/Math.abs(y2 - y1);
                int exposant_fy;
                if (fy != 0)
                    exposant_fy = (int)Math.floor(Math.log(Math.abs(fy))/Math.log(10));
                else
                    exposant_fy = 0;
                double mantisse_fy = fy / (Math.pow(10,exposant_fy));
                mantisse_fy = Math.round(mantisse_fy);
                if (mantisse_fy == 3)
                    mantisse_fy = 4;
                else if (mantisse_fy == 6)
                    mantisse_fy = 5;
                else if (mantisse_fy > 6)
                    mantisse_fy = 10;
                fy = mantisse_fy * Math.pow(10, exposant_fy);
                grilleY(g, fy);
                /*
                int zoomy = 1;
                if (Math.abs(y2 - y1) < 1.0)
                    zoomy = 2; // affiche les diximes
                if (Math.abs(y2 - y1) < 1.0/10)
                    zoomy = 3; // affiche les centimes
                
                if (zoomy == 1)
                    grilleY(g, 1);
                
                if (zoomy == 2)
                    grilleY(g, 10);
                
                if (zoomy == 3)
                    grilleY(g, 100);
                
                g.setFont(boldfont);
                String sy = labely;
                g.drawString(sy, 2, (h - hbottom)/2);
                g.setFont(plafond);
                */
            }
        }
        
        public void grilleX(Graphics g, double facteur) {
            int ix1, ix2, px;
            String scoord;
            if (x2 > x1) {
                ix1 = (int)Math.ceil(x1*facteur);
                ix2 = (int)Math.floor(x2*facteur);
            } else {
                ix1 = (int)Math.ceil(x2*facteur);
                ix2 = (int)Math.floor(x1*facteur);
            }
            for (int i=ix1; i<=ix2; i++) {
                if (x2 > x1)
                    px = (int)Math.rint(((i/facteur - x1) / (x2 - x1)) * (w - wleft)) + wleft;
                else
                    px = w - (int)Math.rint(((i/facteur - x2) / (x1 - x2)) * (w - wleft));
                scoord = joliFormattage(i/facteur, 3);
                g.drawString(scoord, px - (g.getFontMetrics().stringWidth(scoord))/2, h - hbottom/2);
                g.setXORMode(Color.white);
                g.drawLine(px, 0, px, h - hbottom);
                g.setPaintMode();
            }
        }
        
        public void grilleY(Graphics g, double facteur) {
            int iy1, iy2, py;
            String scoord;
            if (y2 > y1) {
                iy1 = (int)Math.ceil(y1*facteur);
                iy2 = (int)Math.floor(y2*facteur);
            } else {
                iy1 = (int)Math.ceil(y2*facteur);
                iy2 = (int)Math.floor(y1*facteur);
            }
            for (int i=iy1; i<=iy2; i++) {
                if (y2 > y1)
                    py = (int)Math.rint(((i/facteur - y1) / (y2 - y1)) * (h - hbottom));
                else
                    py = h - hbottom - (int)Math.rint(((i/facteur - y2) / (y1 - y2)) * (h - hbottom));
                scoord = joliFormattage(i/facteur, 3);
                g.drawString(scoord, 10, py);
                g.setXORMode(Color.white);
                g.drawLine(wleft, py, w, py);
                g.setPaintMode();
            }
        }
        
        public void mouseMoved(MouseEvent e) {
            String slab = "";
            if (e.getX() < w && e.getX() > wleft && e.getY() < h - hbottom) {
                double x = ((e.getX() - wleft) * 1.0 / (w - wleft)) * (x2 - x1) + x1;
                double y = (e.getY() * 1.0 / (h - hbottom)) * (y2 - y1) + y1;
                String sx = joliFormattage(x, 3);
                String sy = joliFormattage(y, 3);
                if (labelx != null && labely != null)
                    slab = " " + sx + ", " + sy;
                else if (labelx != null)
                    slab = " " + sx;
                else if (labely != null)
                    slab = " " + sy;
            }
            /*if (p1 != null && p2 != null) {
                double l = Math.sqrt(Math.pow(p2.x - p1.x, 2) +
                    Math.pow(p2.y - p1.y, 2));
                l = Math.rint(l*10)/10;
                slab += "  + ligne de longueur " + l;
            }*/
            lab.setText(slab);
        }
        
        public void mouseDragged(MouseEvent e) {
            if (p1 != null) {
                if (p2 != null)
                    dessinerLigne();
                p2 = e.getPoint();
                if (p2.x >= w)
                    p2.x = w - 1;
                if (p2.y >= h)
                    p2.y = h - 1;
                dessinerLigne();
                double l = Math.sqrt(Math.pow(p2.x - p1.x, 2) +
                    Math.pow(p2.y - p1.y, 2));
                l = Math.rint(l*10)/10;
                /*String slab = "ligne de longueur " + l;
                lab.setText(slab);*/
            }
        }
        
        public void mouseClicked(MouseEvent e) {
        }
        
        public void mouseEntered(MouseEvent e) {
        }
        
        public void mouseExited(MouseEvent e) {
        }
        
        public void mousePressed(MouseEvent e) {
            if (p2 != null) {
                dessinerLigne();
                p2 = null;
            }
            p1 = e.getPoint();
            if (p1.x >= w)
                p1.x = w - 1;
            if (p1.y >= h)
                p1.y = h - 1;
        }
        
        public void mouseReleased(MouseEvent e) {
            if (p2 != null)
                dessinerLigne();
            p2 = e.getPoint();
            if (p2.x >= w)
                p2.x = w - 1;
            if (p2.y >= h)
                p2.y = h - 1;
            dessinerLigne();
        }
        
        public void keyPressed(KeyEvent e) {
            switch (e.getKeyCode()) {
                case KeyEvent.VK_LEFT: deplacer(-1, 0); break;
                case KeyEvent.VK_RIGHT: deplacer(1, 0); break;
                case KeyEvent.VK_UP: deplacer(0, -1); break;
                case KeyEvent.VK_DOWN: deplacer(0, 1); break;
                default: ;
            }
        }
        
        public void keyReleased(KeyEvent e) {
        }
        
        public void keyTyped(KeyEvent e) {
        }
        
        public void deplacer(int dx, int dy) {
            dessinerObjet();
            if (p1 != null) {
                p1.x += dx;
                p1.y += dy;
            }
            if (p2 != null) {
                p2.x += dx;
                p2.y += dy;
            }
            dessinerObjet();
        }
        
        public void dessinerObjet() {
            if (p1 != null && p2 != null)
                dessinerLigne();
        }
        
        public void dessinerLigne() {
            Graphics g = getGraphics();
            g.setXORMode(Color.white);
            //g.drawLine(p1.x, p1.y, p2.x, p2.y);
            //System.out.println("g.drawLine("+p1.x+", "+p1.y+", "+p2.x+", "+p2.y+");");
            // problme avec Java1.3/Mac si p1.y=p2.y et p1.x>p2.x ou si p1.x=p2.x et p1.y>p2.y
            // (hwaccel)
            if (p1.y == p2.y && p1.x > p2.x)
                g.drawLine(p2.x, p1.y, p1.x, p2.y);
            else if (p1.x == p2.x && p1.y > p2.y)
                g.drawLine(p1.x, p2.y, p2.x, p1.y);
            else
                g.drawLine(p1.x, p1.y, p2.x, p2.y);
            g.setPaintMode();
        }
        
        public Point getP1() {
            Point pim1 = new Point(p1.x - wleft, p1.y);
            return(pim1);
        }
        
        public Point getP2() {
            Point pim2 = new Point(p2.x - wleft, p2.y);
            return(pim2);
        }
        
        public double getFacteurX() {
            return((x2 - x1)/(w - wleft));
        }
        
        public double getFacteurY() {
            return((y2 - y1)/(h - hbottom));
        }
    }

    public void affGraphe(double[] vx, double[] vy) {
        (new FenetreGraphe(vx, vy, labelx, labely)).show();
    }
    
    
    public static String joliFormattage(double d, int ndecimales) {
        int exposant;
        double ddec = Math.pow(10, ndecimales);
        if (d != 0)
            exposant = (int)Math.floor(Math.log(Math.abs(d))/Math.log(10));
        else
            exposant = 0;
        double mantisse = d / (Math.pow(10,exposant));
        String s;
        if (exposant == 0) {
            mantisse = Math.round(mantisse*ddec)/ddec;
            if (mantisse - (int)(mantisse) < 1E-8)
                s = ""+(int)mantisse;
            else
                s = ""+mantisse;
        } else if (exposant == 1) {
            mantisse = Math.round(mantisse*ddec)/(ddec/10);
            if (mantisse - (int)(mantisse) < 1E-8)
                s = ""+(int)mantisse;
            else
                s = ""+mantisse;
        } else if (exposant == 2) {
            mantisse = Math.round(mantisse*ddec)/(ddec/100);
            if (mantisse - (int)(mantisse) < 1E-8)
                s = ""+(int)mantisse;
            else
                s = ""+mantisse;
        } else {
            mantisse = Math.round(mantisse*ddec)/ddec;
            s = mantisse + "E" + exposant; // "E" prend moins de place que ".10^", important en abcisse
        }
        return(s);
    }
    
}
